%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(beam_SUITE).

-export([all/1, packed_registers/1, apply_last/1, apply_last_bif/1]).
-export([applied/2]).

-include("test_server.hrl").

all(suite) -> [packed_registers, apply_last, apply_last_bif].


apply_last(doc) ->
    ["Verifies that apply(M,F,A) is really tail recursive."];
apply_last(suite) -> [];
apply_last(Config) when list(Config) ->
    Pid=spawn(?MODULE, applied, [self(), 10000]),
    Size =
	receive
	    {Pid, finished} ->
		stack_size(Pid)
	after 30000 ->
		?t:fail("applied/2 timed out.")
	end,
    Pid ! die,
    ?t:format("Size: ~p~n", [Size]),
    if
	Size < 700 ->
	    ok;
	true ->
	    ?t:fail("10000 apply() grew stack too much.")
    end,
    ok.

stack_size(Pid) ->
    {heap_size, HS}=process_info(Pid, heap_size),
    {stack_size,SS}=process_info(Pid, stack_size),
    HS+SS.

applied(Starter, 0) ->
    Starter ! {self(), finished},
    receive
	die ->
	    ok
    end,
    ok;
applied(Starter, N) ->
    apply(?MODULE, applied, [Starter, N-1]).

apply_last_bif(doc) ->
    ["Verifies that tail-recursive use of apply(M,F,A) on a Bif works."];
apply_last_bif(suite) -> [];
apply_last_bif(Config) when list(Config) ->
    apply(erlang, abs, [1]).

packed_registers(doc) ->
    "Test three high register numbers in a put_list instruction "
    "(to test if packing works properly).";
packed_registers(suite) -> [];
packed_registers(Config) when list(Config) ->
    ?line PrivDir = ?config(priv_dir, Config),
    ?line Mod = packed_regs,
    ?line Name = filename:join(PrivDir, atom_to_list(Mod) ++ ".erl"),

    %% Generate a module which generates a list of tuples.
    %% put_list(A) -> [{A, 600}, {A, 999}, ... {A, 0}].
    ?line {ok, F} = file:open(Name, [write]),
    ?line ok = io:format(F, "-module(~p).\n", [Mod]),
    ?line ok = io:format(F, "-export([put_list/1]).\n", []),
    ?line ok = io:format(F, "put_list(A) ->\n[", []),
    ?line gen_packed_regs(F, 600),
    ?line file:close(F),

    %% Compile the module.
    ?line io:format("Compiling: ~s\n", [Name]),
    ?line CompRc = compile:file(Name, [{outdir, PrivDir}, report, time]),
    ?line io:format("Result: ~p\n",[CompRc]),
    ?line {ok, Mod} = CompRc,

    %% Load it.
    ?line io:format("Loading...\n",[]),
    ?line LoadRc = code:load_abs(filename:join(PrivDir, atom_to_list(Mod))),
    ?line {module, Module} = LoadRc,

    %% Call it and verify result.
    ?line Term = {a, b},
    ?line L = Mod:put_list(Term),
    ?line verify_packed_regs(L, Term, 600),
    ok.

gen_packed_regs(F, 0) ->
    ok = io:format(F, "{A, ~w}].\n", [0]);
gen_packed_regs(F, N) ->
    ok = io:format(F, "{A, ~w},\n", [N]),
    gen_packed_regs(F, N-1).

verify_packed_regs([], _, -1) -> ok;
verify_packed_regs([{Term, N}| T], Term, N) ->
    verify_packed_regs(T, Term, N-1);
verify_packed_regs(L, Term, N) ->
    ?line ok = io:format("Expected [{~p, ~p}|T]; got\n~p\n", [Term, N, L]),
    ?line test_server:fail().

