%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(statistics_SUITE).

%% Tests the statistics/1 bif.

-export([all/1,
	 init_per_testcase/2,
	 fin_per_testcase/2,
	 wall_clock/1, wall_clock_zero_diff/1, wall_clock_update/1,
	 runtime/1, runtime_zero_diff/1, runtime_zero_update/1,
	 runtime_update/1, runtime_diff/1,
	 run_queue/1, run_queue_one/1,
	 reductions/1, reductions_big/1, garbage_collection/1, io/1,
	 badarg/1]).

%% Internal exports.

-export([hog/1]).

-include("test_server.hrl").

init_per_testcase(Case, Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(300)),
    [{watchdog, Dog}|Config].

fin_per_testcase(Case, Config) ->
    Dog = ?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),
    ok.

all(suite) -> [wall_clock, runtime, reductions, reductions_big, run_queue,
	       garbage_collection, io, badarg].


%%% Testing statistics(wall_clock).

wall_clock(suite) -> [wall_clock_zero_diff, wall_clock_update].


wall_clock_zero_diff(doc) ->
    "Tests that the 'Wall clock since last call' element of the result "
    "is zero when statistics(runtime) is called twice in succession.";
wall_clock_zero_diff(suite) -> [];
wall_clock_zero_diff(Config) when list(Config) ->
    wall_clock_zero_diff1(16).

wall_clock_zero_diff1(N) when N > 0 ->
    ?line {Time, _} = statistics(wall_clock),
    ?line case statistics(wall_clock) of
	      {Time, 0} -> ok;
	      _ -> wall_clock_zero_diff1(N-1)
    end;
wall_clock_zero_diff1(0) ->
    ?line test_server:fail("Difference never zero.").

wall_clock_update(doc) ->
    "Test that the time differences returned by two calls to "
    "statistics(wall_clock) are compatible, and are within a small number "
    "of ms of the amount of real time we waited for.";
wall_clock_update(suite) -> [];
wall_clock_update(Config) when list(Config) ->
    wall_clock_update1(6).

wall_clock_update1(N) when N > 0 ->
    ?line {T1_wc_time, _} = statistics(wall_clock),
    ?line receive after 1000 -> ok end,
    ?line {T2_wc_time, Wc_Diff} = statistics(wall_clock),

    ?line Wc_Diff = T2_wc_time - T1_wc_time,
    ?line test_server:format("Wall clock diff = ~p; should be  = 1000..1040~n",
			     [Wc_Diff]),
    ?line Wc_Diff =< 1040,
    ?line Wc_Diff >= 1000,
    wall_clock_update1(N-1);
wall_clock_update1(0) ->
    ok.


%%% Test statistics(runtime).

runtime(suite) -> [runtime_zero_diff, runtime_zero_update, runtime_update,
		   runtime_diff].

runtime_zero_diff(doc) ->
    "Tests that the difference between the times returned from two succesive "
	"calls to statistics(runtime) is zero.";
runtime_zero_diff(suite) -> [];
runtime_zero_diff(Config) when list(Config) ->
    ?line runtime_zero_diff1(16).

runtime_zero_diff1(N) when N > 0 ->
    ?line {T1, _} = statistics(runtime),
    ?line case statistics(runtime) of
	      {T1, 0} -> ok;
	      _ -> runtime_zero_diff1(N-1)
	  end;
runtime_zero_diff1(0) ->
    ?line test_server:fail("statistics(runtime) never returned zero difference").

runtime_zero_update(doc) ->
    "Test that the time differences returned by two calls to "
    "statistics(runtime) one second apart is zero.";
runtime_zero_update(suite) -> [];
runtime_zero_update(Config) when list(Config) ->
    ?line runtime_zero_update1(6).

runtime_zero_update1(N) when N > 0 ->
    ?line {T1, _} = statistics(runtime),
    ?line receive after 10000 -> ok end,
    ?line case statistics(runtime) of
	      {T, Td} when Td =< 80 ->
		  test_server:format("ok, Runtime before: {~p, _} after: {~p, ~p}",
				     [T1, T, Td]),
		  ok;
	      {T, R} ->
		  test_server:format("nok, Runtime before: {~p, _} after: {~p, ~p}", 
				     [T1, T, R]),
		  runtime_zero_update1(N-1)
	  end;
runtime_zero_update1(0) ->
    ?line test_server:fail("statistics(runtime) never returned zero difference").

runtime_update(doc) ->
    "Test that the statistics(runtime) returns a substanstially updated difference "
    "after running a process that takes all CPU power of the Erlang process "
    "for a second.";
runtime_update(suite) -> [];
runtime_update(Config) when list(Config) ->
    ?line process_flag(priority, high),
    ?line test_server:m_out_of_n(1, 10, fun runtime_update/0).

runtime_update() ->
    ?line {T1, _} = block_cpu(),
    ?line {T2, Diff} = statistics(runtime),
    ?line case {T2-T1, abs(Diff-1000)} of
	      {Diff, Delta} when Delta =< 50 ->
		  ok;
	      {Diff, Delta} ->
		  test_server:format("T1 = ~p, T2 = ~p, Diff = ~p, abs(Diff-1000) = ~p",
				     [T1, T2, Diff, Delta])
	  end.
    
%% Blocks the CPU for about one second.  Returns the result from
%% statistics(runtime) before the second started.

block_cpu() ->
    ?line {_, _, Initial_Sec} = time(),
    ?line block_cpu(Initial_Sec).

block_cpu(Sec) ->
    ?line case time() of
	      {_, _, Sec} ->
		  block_cpu(Sec);
	      {_, _, New_Sec} ->
		  block_cpu(New_Sec, statistics(runtime))
	  end.

block_cpu(Sec, Before) ->
    ?line case time() of
	      {_, _, Sec} ->
		  block_cpu(Sec, Before);
	      _ ->
		  Before
	  end.

runtime_diff(doc) ->
    "Test that the difference between two consecutive absolute runtimes is "
    "equal to the last relative runtime. The loop runs a lot of times since "
    "the bug which this test case tests for showed up only rarely.";
runtime_diff(suite) -> [];
runtime_diff(Config) when list(Config) ->
    runtime_diff1(1000).

runtime_diff1(N) when N > 0 ->
    ?line {T1_wc_time, _} = statistics(runtime),
    ?line do_much(),
    ?line {T2_wc_time, Wc_Diff} = statistics(runtime),
    ?line Wc_Diff = T2_wc_time - T1_wc_time,
    runtime_diff1(N-1);
runtime_diff1(0) ->
    ok.

%%% do_much(100000) takes about 760 ms on boromir.
%%% do_much(1000) takes about 8 ms on boromir.

do_much() ->
    do_much(1000).

do_much(0) ->
    ok;
do_much(N) ->
    X = 4784728478274827 * 72874284728472,
    do_much(N-1).


reductions(doc) ->
    "Test that statistics(reductions) is callable, and that "
	"Total_Reductions and Reductions_Since_Last_Call make sense. "
	"(This to fail on pre-R3A version of JAM.";
reductions(suite) ->
    [];
reductions(Config) when list(Config) ->
    {Reductions, _} = statistics(reductions),

    %% Each loop of reductions/2 takes 4 reductions + that the garbage built
    %% outside the heap in the BIF calls will bump the reductions.
    %% 300 * 4 is more than CONTEXT_REDS (1000).  Thus, there will be one or
    %% more context switches.

    reductions(300, Reductions).

reductions(N, Previous) when N > 0 ->
    ?line {Reductions, Diff} = statistics(reductions),
    ?line build_some_garbage(),
    ?line if Reductions > 0 -> ok end,
    ?line if Diff >= 0 -> ok end,
    io:format("Previous = ~p, Reductions = ~p, Diff = ~p, DiffShouldBe = ~p",
	      [Previous, Reductions, Diff, Reductions-Previous]),
    ?line if Reductions == Previous+Diff -> reductions(N-1, Reductions) end;
reductions(0, _) ->
    ok.

build_some_garbage() ->
    %% This will build garbage outside the process heap, which will cause
    %% a garbage collection in the scheduler.
    processes().

reductions_big(doc) ->
    "Test that the number of reductions can be returned as a big number.";
reductions_big(suite) ->
    {req, [{time, 300}]};
reductions_big(Config) when list(Config) ->
    ?line reductions_big_loop(),
    ok.

reductions_big_loop() ->
    erlang:yield(),
    case statistics(reductions) of
	{Red, Diff} when Red >= 16#7ffFFFF ->
	    ok = io:format("Reductions = ~w, Diff = ~w", [Red, Diff]);
	_ ->
	    reductions_big_loop()
    end.


%%% Tests of statistics(run_queue).

run_queue(suite) -> [run_queue_one].

run_queue_one(doc) ->
    "Tests that statistics(run_queue) returns 1 if we start a "
    "CPU-bound process.";
run_queue_one(suite) -> [];
run_queue_one(Config) when list(Config) ->
    ?line Hog = spawn_link(?MODULE, hog, [self()]),
    ?line receive
	hog_started ->
	    Hog ! go
    end,
    ?line receive after 100 -> ok end,		% Give hog a head start.
    ?line case statistics(run_queue) of
	      N when N >= 1 -> ok;
	      Other -> ?line ?t:fail({unexpected,Other})
	  end,
    ok.

%% CPU-bound process, going at low priority.  It will always be ready
%% to run.
    
hog(Pid) ->
    ?line process_flag(priority, low),
    ?line Pid ! hog_started,
    ?line receive
	go -> hog_iter(0)
    end.

hog_iter(N) when N > 0 ->
    ?line hog_iter(N-1);
hog_iter(0) ->
    ?line hog_iter(10000).


garbage_collection(doc) ->
    "Tests that statistics(garbage_collection) is callable. "
    "It is not clear how to test anything more.";
garbage_collection(suite) -> [];
garbage_collection(Config) when list(Config) ->
    ?line case statistics(garbage_collection) of
	      {I1, I2, 0} when integer(I1), integer(I2) -> ok
	  end.

io(doc) ->
    "Tests that statistics(io) is callable. "
    "This could be improved to test something more.";
io(suite) -> [];
io(Config) when list(Config) ->
    ?line case statistics(io) of
	      {{input, In}, {output, Out}} when integer(In), integer(Out) -> ok
	  end.

badarg(doc) ->
    "Tests that some illegal arguments to statistics fails.";
badarg(suite) -> [];
badarg(Config) when list(Config) ->
    ?line case catch statistics(1) of
	      {'EXIT', {badarg, _}} -> ok
	  end,
    ?line case catch statistics(bad_atom) of
	      {'EXIT', {badarg, _}} -> ok
	  end.
